@@ -1,7 +1,7 @@
+from typing import TypeVar, Set
 import logging
 import weakref
 import trio
-import gc
 import functools
 import threading
 
@@ -12,6 +12,19 @@ from .guard import exclusive
 from ..debug import NameRepr
 
 
+T = TypeVar("T")
+
+
+def _weakset_copy(ws: "weakref.WeakSet[T]") -> Set[T]:
+    """Thread-safely copy all items from a weakset to a set"""
+    # The various WeakSet methods are not thread-safe because they miss locking.
+    # The main issue is that all copy approaches use ``__iter__``, which is not
+    # thread-safe against items being garbage collected. However, we can access
+    # the actual backing real set ``ws.data`` and ``set(some_set)`` is GIL-atomic.
+    refs = set(ws.data)
+    return {item for item in (ref() for ref in refs) if item is not None}
+
+
 class ServiceUnit(object):
     """
     Definition for running a service
@@ -35,9 +48,9 @@ class ServiceUnit(object):
         ServiceUnit.__active_units__.add(self)
 
     @classmethod
-    def units(cls):
-        """Sequence of all currently defined units"""
-        return list(cls.__active_units__)
+    def units(cls) -> "Set[ServiceUnit]":
+        """Container of all currently defined units"""
+        return _weakset_copy(cls.__active_units__)
 
     @property
     def running(self):
@@ -96,6 +109,16 @@ def service(flavour):
 class ServiceRunner(object):
     """
     Runner for coroutines, subroutines and services
+
+    The service runner prevents silent failures by tracking concurrent tasks
+    and therefore provides safer concurrency.
+    If any task fails with an exception or provides
+    unexpected output values, this is registered as an error; the runner will
+    gracefully shut down all tasks in this case.
+
+    To provide ``async`` concurrency, the runner also manages common
+    ``async`` event loops and tracks them for failures as well. As a result,
+    ``async`` code should usually use the "current" event loop directly.
     """
 
     def __init__(self, accept_delay: float = 1):
@@ -103,6 +126,7 @@ class ServiceRunner(object):
         self._meta_runner = MetaRunner()
         self._must_shutdown = False
         self._is_shutdown = threading.Event()
+        self._is_shutdown.set()
         self.running = threading.Event()
         self.accept_delay = accept_delay
 
@@ -136,14 +160,8 @@ class ServiceRunner(object):
         Since services are globally defined, only one :py:class:`ServiceRunner`
         may :py:meth:`accept` payloads at any time.
         """
-        if self._meta_runner:
-            raise RuntimeError("payloads scheduled for %s before being started" % self)
         self._must_shutdown = False
         self._logger.info("%s starting", self.__class__.__name__)
-        # force collecting objects so that defunct,
-        # migrated and overwritten services are destroyed now
-        gc.collect()
-        self._adopt_services()
         self.adopt(self._accept_services, flavour=trio)
         self._meta_runner.run()
 
@@ -163,7 +181,9 @@ class ServiceRunner(object):
                 self._adopt_services()
                 await trio.sleep(delay)
                 delay = min(delay + increase, max_delay)
-        except Exception:
+        except trio.Cancelled:
+            self._logger.info("%s cancelled", self.__class__.__name__)
+        except BaseException:
             self._logger.exception("%s aborted", self.__class__.__name__)
             raise
         else:
